@@ -1,6 +1,10 @@
+# frozen_string_literal: true
+
 require "open3"
 require "optparse"
 require "shellwords"
+require "abbrev"
+require "stringio"
 
 module RBS
   class CLI
@@ -17,7 +21,7 @@ module RBS
 
         @libs = []
         @dirs = []
-        @config_path = Collection::Config::PATH
+        @config_path = Collection::Config.find_config_path || Collection::Config::PATH
       end
 
       def loader
@@ -27,7 +31,12 @@ module RBS
         end
 
         loader = EnvironmentLoader.new(core_root: core_root, repository: repository)
-        lock = config_path&.then { |p| Collection::Config.lockfile_of(p) }
+        if config_path
+          lock_path = Collection::Config.to_lockfile_path(config_path)
+          if lock_path.file?
+            lock = Collection::Config::Lockfile.from_lockfile(lockfile_path: lock_path, data: YAML.load_file(lock_path.to_s))
+          end
+        end
         loader.add_collection(lock) if lock
 
         dirs.each do |dir|
@@ -56,11 +65,11 @@ module RBS
           self.core_root = nil
         end
 
-        opts.on('--collection PATH', "File path of collection configration (default: #{Collection::Config::PATH})") do |path|
+        opts.on('--collection PATH', "File path of collection configuration (default: #{@config_path})") do |path|
           self.config_path = Pathname(path).expand_path
         end
 
-        opts.on('--no-collection', 'Ignore collection configration') do
+        opts.on('--no-collection', 'Ignore collection configuration') do
           self.config_path = nil
         end
 
@@ -74,13 +83,14 @@ module RBS
 
     attr_reader :stdout
     attr_reader :stderr
+    attr_reader :original_args
 
     def initialize(stdout:, stderr:)
       @stdout = stdout
       @stderr = stderr
     end
 
-    COMMANDS = [:ast, :annotate, :list, :ancestors, :methods, :method, :validate, :constant, :paths, :prototype, :vendor, :parse, :test, :collection]
+    COMMANDS = [:ast, :annotate, :list, :ancestors, :methods, :method, :validate, :constant, :paths, :prototype, :vendor, :parse, :test, :collection, :subtract]
 
     def parse_logging_options(opts)
       opts.on("--log-level LEVEL", "Specify log level (defaults to `warn`)") do |level|
@@ -100,6 +110,8 @@ module RBS
     end
 
     def run(args)
+      @original_args = args.dup
+
       options = LibraryOptions.new
 
       opts = OptionParser.new
@@ -220,6 +232,19 @@ EOB
             end
           end
         end
+
+        env.class_alias_decls.each do |name, entry|
+          case entry
+          when Environment::ModuleAliasEntry
+            if list.include?(:module)
+              stdout.puts "#{name} (module alias)"
+            end
+          when Environment::ClassAliasEntry
+            if list.include?(:class)
+              stdout.puts "#{name} (class alias)"
+            end
+          end
+        end
       end
 
       if list.include?(:interface)
@@ -250,6 +275,11 @@ EOU
         opts.on("--singleton", "Ancestors of singleton of the given type_name") { kind = :singleton }
       end.order!(args)
 
+      unless args.size == 1
+        stdout.puts "Expected one argument."
+        return
+      end
+
       loader = options.loader()
 
       env = Environment.from_loader(loader).resolve_type_names
@@ -257,7 +287,10 @@ EOU
       builder = DefinitionBuilder::AncestorBuilder.new(env: env)
       type_name = TypeName(args[0]).absolute!
 
-      if env.class_decls.key?(type_name)
+      case env.constant_entry(type_name)
+      when Environment::ClassEntry, Environment::ModuleEntry, Environment::ClassAliasEntry, Environment::ModuleAliasEntry
+        type_name = env.normalize_module_name(type_name)
+
         ancestors = case kind
                     when :instance
                       builder.instance_ancestors(type_name)
@@ -319,7 +352,7 @@ EOU
       builder = DefinitionBuilder.new(env: env)
       type_name = TypeName(args[0]).absolute!
 
-      if env.class_decls.key?(type_name)
+      if env.module_name?(type_name)
         definition = case kind
                      when :instance
                        builder.build_instance(type_name)
@@ -373,7 +406,7 @@ EOU
       type_name = TypeName(args[0]).absolute!
       method_name = args[1].to_sym
 
-      unless env.class_decls.key?(type_name)
+      unless env.module_name?(type_name)
         stdout.puts "Cannot find class: #{type_name}"
         return
       end
@@ -419,7 +452,6 @@ Examples:
 EOU
 
         opts.on("--silent") do
-          require "stringio"
           @stdout = StringIO.new
         end
       end.parse!(args)
@@ -428,15 +460,15 @@ EOU
       env = Environment.from_loader(loader).resolve_type_names
 
       builder = DefinitionBuilder.new(env: env)
-      validator = Validator.new(env: env, resolver: TypeNameResolver.from_env(env))
+      validator = Validator.new(env: env, resolver: Resolver::TypeNameResolver.new(env))
 
       env.class_decls.each do |name, decl|
         stdout.puts "Validating class/module definition: `#{name}`..."
         builder.build_instance(name).each_type do |type|
-          validator.validate_type type, context: [Namespace.root]
+          validator.validate_type type, context: nil
         end
         builder.build_singleton(name).each_type do |type|
-          validator.validate_type type, context: [Namespace.root]
+          validator.validate_type type, context: nil
         end
 
         d = decl.primary.decl
@@ -457,10 +489,15 @@ EOU
         end
       end
 
+      env.class_alias_decls.each do |name, entry|
+        stdout.puts "Validating class/module alias definition: `#{name}`..."
+        validator.validate_class_alias(entry: entry)
+      end
+
       env.interface_decls.each do |name, decl|
         stdout.puts "Validating interface: `#{name}`..."
         builder.build_interface(name).each_type do |type|
-          validator.validate_type type, context: [Namespace.root]
+          validator.validate_type type, context: nil
         end
 
         validator.validate_type_params(
@@ -485,13 +522,13 @@ EOU
 
       env.global_decls.each do |name, global|
         stdout.puts "Validating global: `#{name}`..."
-        validator.validate_type global.decl.type, context: [Namespace.root]
+        validator.validate_type global.decl.type, context: nil
       end
 
-      env.alias_decls.each do |name, decl|
+      env.type_alias_decls.each do |name, decl|
         stdout.puts "Validating alias: `#{name}`..."
         builder.expand_alias1(name).tap do |type|
-          validator.validate_type type, context: [Namespace.root]
+          validator.validate_type type, context: nil
         end
         validator.validate_type_alias(entry: decl)
       end
@@ -527,14 +564,26 @@ EOU
       env = Environment.from_loader(loader).resolve_type_names
 
       builder = DefinitionBuilder.new(env: env)
-      table = ConstantTable.new(builder: builder)
+      resolver = Resolver::ConstantResolver.new(builder: builder)
 
-      namespace = context ? Namespace.parse(context).absolute! : Namespace.root
-      stdout.puts "Context: #{namespace}"
-      name = Namespace.parse(args[0]).to_type_name
-      stdout.puts "Constant name: #{name}"
+      resolver_context = context ? [nil, TypeName(context).absolute!] : nil #: Resolver::context
+      stdout.puts "Context: #{context}"
+      const_name = TypeName(args[0])
+      stdout.puts "Constant name: #{const_name}"
 
-      constant = table.resolve_constant_reference(name, context: namespace.ascend.to_a)
+      if const_name.absolute?
+        constant = resolver.table.constant(const_name)
+      else
+        head, *components = const_name.to_namespace.path
+        head or raise
+
+        constant = resolver.resolve(head, context: resolver_context)
+        constant = components.inject(constant) do |const, component|
+          if const
+            resolver.resolve_child(const.name, component)
+          end
+        end
+      end
 
       if constant
         stdout.puts " => #{constant.name}: #{constant.type}"
@@ -590,7 +639,7 @@ EOU
 
       case format
       when "rbi", "rb"
-        decls = run_prototype_file(format, args)
+        run_prototype_file(format, args)
       when "runtime"
         require_libs = []
         relative_libs = []
@@ -638,6 +687,9 @@ EOU
         end
 
         decls = Prototype::Runtime.new(patterns: args, env: env, merge: merge, owners_included: owners_included).decls
+
+        writer = Writer.new(out: stdout)
+        writer.write decls
       else
         stdout.puts <<EOU
 Usage: rbs prototype [generator...] [args...]
@@ -653,13 +705,6 @@ Examples:
 EOU
         exit 1
       end
-
-      if decls
-        writer = Writer.new(out: stdout)
-        writer.write decls
-      else
-        exit 1
-      end
     end
 
     def run_prototype_file(format, args)
@@ -667,6 +712,13 @@ EOU
                        "\n** This command does not work on this interpreter (#{RUBY_ENGINE}) **\n"
                      end
 
+      # @type var output_dir: Pathname?
+      output_dir = nil
+      # @type var base_dir: Pathname?
+      base_dir = nil
+      # @type var force: bool
+      force = false
+
       opts = OptionParser.new
       opts.banner = <<EOU
 Usage: rbs prototype #{format} [files...]
@@ -680,7 +732,25 @@ Examples:
 
   $ rbs prototype rb lib/foo.rb
   $ rbs prototype rbi sorbet/rbi/foo.rbi
+
+You can run the tool in *batch* mode by passing `--out-dir` option.
+
+  $ rbs prototype rb --out-dir=sig lib/foo.rb
+  $ rbs prototype rbi --out-dir=sig/models --base-dir=app/models app/models
 EOU
+
+      opts.on("--out-dir=DIR", "Specify the path to save the generated RBS files") do |path|
+        output_dir = Pathname(path)
+      end
+
+      opts.on("--base-dir=DIR", "Specify the path to calculate the relative path to save the generated RBS files") do |path|
+        base_dir = Pathname(path)
+      end
+
+      opts.on("--force", "Overwrite existing RBS files") do
+        force = true
+      end
+
       opts.parse!(args)
 
       unless has_parser?
@@ -693,18 +763,108 @@ EOU
         return nil
       end
 
-      parser = case format
-               when "rbi"
-                 Prototype::RBI.new()
-               when "rb"
-                 Prototype::RB.new()
-               end
-
-      args.each do |file|
-        parser.parse Pathname(file).read
+      new_parser = -> do
+        case format
+        when "rbi"
+          Prototype::RBI.new()
+        when "rb"
+          Prototype::RB.new()
+        else
+          raise
+        end
       end
 
-      parser.decls
+      input_paths = args.map {|arg| Pathname(arg) }
+
+      if output_dir
+        # @type var skip_paths: Array[Pathname]
+        skip_paths = []
+
+        # batch mode
+        input_paths.each do |path|
+          stdout.puts "Processing `#{path}`..."
+          ruby_files =
+            if path.file?
+              [path]
+            else
+              path.glob("**/*.rb").sort
+            end
+
+          ruby_files.each do |file_path|
+            stdout.puts "  Generating RBS for `#{file_path}`..."
+
+            relative_path =
+              if base_dir
+                file_path.relative_path_from(base_dir)
+              else
+                if top = file_path.descend.first
+                  case
+                  when top == Pathname("lib")
+                    file_path.relative_path_from(top)
+                  when top == Pathname("app")
+                    file_path.relative_path_from(top)
+                  else
+                    file_path
+                  end
+                else
+                  file_path
+                end
+              end
+            relative_path = relative_path.cleanpath()
+
+            if relative_path.absolute? || relative_path.descend.first&.to_s == ".."
+              stdout.puts "  ⚠️  Cannot write the RBS to outside of the output dir: `#{relative_path}`"
+              next
+            end
+
+            output_path = (output_dir + relative_path).sub_ext(".rbs")
+
+            parser = new_parser[]
+            parser.parse file_path.read()
+
+            if output_path.file?
+              if force
+                stdout.puts "    - Writing RBS to existing file `#{output_path}`..."
+              else
+                stdout.puts "    - Skipping existing file `#{output_path}`..."
+                skip_paths << file_path
+                next
+              end
+            else
+              stdout.puts "    - Writing RBS to `#{output_path}`..."
+            end
+
+            (output_path.parent).mkpath
+            output_path.open("w") do |io|
+              writer = Writer.new(out: io)
+              writer.write(parser.decls)
+            end
+          end
+        end
+
+        unless skip_paths.empty?
+          stdout.puts
+          stdout.puts ">>>> Skipped existing #{skip_paths.size} files. Use `--force` option to update the files."
+          command = original_args.take(original_args.size - input_paths.size)
+
+          skip_paths.take(10).each do |path|
+            stdout.puts "  #{defined?(Bundler) ? "bundle exec " : ""}rbs #{Shellwords.join(command)} --force #{Shellwords.escape(path.to_s)}"
+          end
+          if skip_paths.size > 10
+            stdout.puts "  ..."
+          end
+        end
+      else
+        # file mode
+        parser = new_parser[]
+
+        input_paths.each do |file|
+          parser.parse file.read()
+        end
+
+        writer = Writer.new(out: stdout)
+        writer.write parser.decls
+      end
     end
 
     def run_vendor(args, options)
@@ -761,6 +921,10 @@ Options:
     end
 
     def run_parse(args, options)
+      parse_method = :parse_signature
+      # @type var e_code: String?
+      e_code = nil
+
       OptionParser.new do |opts|
         opts.banner = <<-EOB
 Usage: rbs parse [files...]
@@ -770,23 +934,36 @@ Parse given RBS files and print syntax errors.
 Examples:
 
   $ rbs parse sig/app/models.rbs sig/app/controllers.rbs
+
+Options:
         EOB
+
+        opts.on('-e CODE', 'One line RBS script to parse') { |e| e_code = e }
+        opts.on('--type', 'Parse code as a type') { |e| parse_method = :parse_type }
+        opts.on('--method-type', 'Parse code as a method type') { |e| parse_method = :parse_method_type }
       end.parse!(args)
 
-      loader = options.loader()
-
       syntax_error = false
-      args.each do |path|
+      bufs = args.flat_map do |path|
         path = Pathname(path)
-        loader.each_file(path, skip_hidden: false, immediate: true) do |file_path|
-          RBS.logger.info "Parsing #{file_path}..."
-          buffer = Buffer.new(content: file_path.read, name: file_path)
-          Parser.parse_signature(buffer)
-        rescue RBS::ParsingError => ex
-          stdout.puts ex.message
-          syntax_error = true
+        FileFinder.each_file(path, skip_hidden: false, immediate: true).map do |file_path|
+          Buffer.new(content: file_path.read, name: file_path)
         end
       end
+      bufs << Buffer.new(content: e_code, name: '-e') if e_code
+
+      bufs.each do |buf|
+        RBS.logger.info "Parsing #{buf.name}..."
+        case parse_method
+        when :parse_signature
+          Parser.parse_signature(buf)
+        else
+          Parser.public_send(parse_method, buf, require_eof: true)
+        end
+      rescue RBS::ParsingError => ex
+        stdout.puts ex.message
+        syntax_error = true
+      end
 
       exit 1 if syntax_error
     end
@@ -896,7 +1073,7 @@ EOB
       env_hash = {
         'RUBYOPT' => "#{ENV['RUBYOPT']} -rrbs/test/setup",
         'RBS_TEST_OPT' => test_opt(options),
-        'RBS_TEST_LOGLEVEL' => RBS.logger_level,
+        'RBS_TEST_LOGLEVEL' => %w(DEBUG INFO WARN ERROR FATAL)[RBS.logger_level || 5] || "UNKNOWN",
         'RBS_TEST_SAMPLE_SIZE' => sample_size,
         'RBS_TEST_DOUBLE_SUITE' => double_suite,
         'RBS_TEST_UNCHECKED_CLASSES' => (unchecked_classes.join(',') unless unchecked_classes.empty?),
@@ -919,15 +1096,16 @@ EOB
       config_path = options.config_path or raise
       lock_path = Collection::Config.to_lockfile_path(config_path)
 
-      case args[0]
+      subcommand = Abbrev.abbrev(['install', 'update', 'help'])[args[0]] || args[0]
+      case subcommand
       when 'install'
         unless params[:frozen]
-          Collection::Config.generate_lockfile(config_path: config_path, gemfile_lock_path: Pathname('./Gemfile.lock'))
+          Collection::Config.generate_lockfile(config_path: config_path, definition: Bundler.definition)
         end
         Collection::Installer.new(lockfile_path: lock_path, stdout: stdout).install_from_lockfile
       when 'update'
         # TODO: Be aware of argv to update only specified gem
-        Collection::Config.generate_lockfile(config_path: config_path, gemfile_lock_path: Pathname('./Gemfile.lock'), with_lockfile: false)
+        Collection::Config.generate_lockfile(config_path: config_path, definition: Bundler.definition, with_lockfile: false)
         Collection::Installer.new(lockfile_path: lock_path, stdout: stdout).install_from_lockfile
       when 'init'
         if config_path.exist?
@@ -938,11 +1116,16 @@ EOB
         config_path.write(<<~'YAML')
           # Download sources
           sources:
-            - name: ruby/gem_rbs_collection
+            - type: git
+              name: ruby/gem_rbs_collection
               remote: https://github.com/ruby/gem_rbs_collection.git
               revision: main
               repo_dir: gems
 
+          # You can specify local directories as sources also.
+          # - type: local
+          #   path: path/to/your/local/repository
+
           # A directory to install the downloaded RBSs
           path: .gem_rbs_collection
 
@@ -976,7 +1159,7 @@ EOB
 
           Examples:
 
-            # Initialize the configration file
+            # Initialize the configuration file
             $ rbs collection init
 
             # Generate the lock file and install RBSs from the lock file
@@ -984,9 +1167,86 @@ EOB
 
             # Update the RBSs
             $ rbs collection update
+
+          Options:
         HELP
         opts.on('--frozen') if args[0] == 'install'
       end
     end
+
+    def run_subtract(args, _)
+      write_to_file = false
+      # @type var subtrahend_paths: Array[String]
+      subtrahend_paths = []
+
+      opts = OptionParser.new do |opts|
+        opts.banner = <<~HELP
+          Usage:
+            rbs subtract [options...] minuend.rbs [minuend2.rbs, ...] subtrahend.rbs
+            rbs subtract [options...] minuend.rbs [minuend2.rbs, ...] --subtrahend subtrahend_1.rbs --subtrahend subtrahend_2.rbs
+
+          Remove duplications between RBS files.
+
+          Examples:
+
+            # Generate RBS files from the codebase.
+            $ rbs prototype rb lib/ > generated.rbs
+
+            # Write more descrictive types by hand.
+            $ $EDITOR handwritten.rbs
+
+            # Remove hand-written method definitions from generated.rbs.
+            $ rbs subtract --write generated.rbs handwritten.rbs
+
+          Options:
+        HELP
+        opts.on('-w', '--write', 'Overwrite files directry') { write_to_file = true }
+        opts.on('--subtrahend=PATH', '') { |path| subtrahend_paths << path }
+        opts.parse!(args)
+      end
+
+      if subtrahend_paths.empty?
+        *minuend_paths, subtrahend_path = args
+        unless subtrahend_path
+          stdout.puts opts.help
+          exit 1
+        end
+        subtrahend_paths << subtrahend_path
+      else
+        minuend_paths = args
+      end
+
+      if minuend_paths.empty?
+        stdout.puts opts.help
+        exit 1
+      end
+
+      subtrahend = Environment.new.tap do |env|
+        loader = EnvironmentLoader.new(core_root: nil)
+        subtrahend_paths.each do |path|
+          loader.add(path: Pathname(path))
+        end
+        loader.load(env: env)
+      end
+
+      minuend_paths.each do |minuend_path|
+        FileFinder.each_file(Pathname(minuend_path), immediate: true, skip_hidden: true) do |rbs_path|
+          buf = Buffer.new(name: rbs_path, content: rbs_path.read)
+          _, dirs, decls = Parser.parse_signature(buf)
+          subtracted = Subtractor.new(decls, subtrahend).call
+
+          io = StringIO.new
+          w = Writer.new(out: io)
+          w.write(dirs)
+          w.write(subtracted)
+
+          if write_to_file
+            rbs_path.write(io.string)
+          else
+            stdout.puts(io.string)
+          end
+        end
+      end
+    end
   end
 end
