@@ -1,4 +1,14 @@
-# Thread and Fiber
+show_limit %q{
+  threads = []
+  begin
+    threads << Thread.new{sleep}
+
+    raise Exception, "skipping" if threads.count >= 10_000
+  rescue Exception => error
+    puts "Thread count: #{threads.count} (#{error})"
+    break
+  end while true
+} if false # disable to pass CI
 
 assert_equal %q{ok}, %q{
   Thread.new{
@@ -36,7 +46,7 @@ begin
     }
   }
 rescue ThreadError => e
-  :ok if /can't create Thread/ =~ e.message
+  /can't create Thread/ =~ e.message ? :ok : e.message
 end
 }
 assert_equal %q{ok}, %q{
@@ -50,7 +60,7 @@ begin
     }
   }
 rescue ThreadError => e
-  :ok if /can't create Thread/ =~ e.message
+  /can't create Thread/ =~ e.message ? :ok : e.message
 end
 }
 assert_equal %q{ok}, %q{
@@ -233,7 +243,7 @@ assert_equal 'true', %{
 }
 
 assert_equal 'ok', %{
-  open("zzz.rb", "w") do |f|
+  open("zzz_t1.rb", "w") do |f|
     f.puts <<-END
       begin
         Thread.new { fork { GC.start } }.join
@@ -244,7 +254,7 @@ assert_equal 'ok', %{
       end
     END
   end
-  require "./zzz.rb"
+  require "./zzz_t1.rb"
   $result
 }
 
@@ -298,10 +308,6 @@ assert_equal 'ok', %q{
   m
 }, '[ruby-dev:34492]'
 
-assert_normal_exit %q{
-  at_exit { Fiber.new{}.resume }
-}
-
 assert_normal_exit %q{
   g = enum_for(:local_variables)
   loop { g.next }
@@ -327,10 +333,6 @@ assert_normal_exit %q{
   loop { g.next }
 }, '[ruby-dev:34128]'
 
-assert_normal_exit %q{
-  Fiber.new(&Object.method(:class_eval)).resume("foo")
-}, '[ruby-dev:34128]'
-
 assert_normal_exit %q{
   Thread.new("foo", &Object.method(:class_eval)).join
 }, '[ruby-dev:34128]'
@@ -347,7 +349,7 @@ assert_equal 'ok', %q{
 
 assert_equal 'ok', %q{
   begin
-    m1, m2 = Mutex.new, Mutex.new
+    m1, m2 = Thread::Mutex.new, Thread::Mutex.new
     f1 = f2 = false
     Thread.new { m1.lock; f2 = true; sleep 0.001 until f1; m2.lock }
     m2.lock; f1 = true; sleep 0.001 until f2; m1.lock
@@ -358,32 +360,32 @@ assert_equal 'ok', %q{
 }
 
 assert_equal 'ok', %q{
-  m = Mutex.new
+  m = Thread::Mutex.new
   Thread.new { m.lock }; sleep 0.1; m.lock
   :ok
 }
 
 assert_equal 'ok', %q{
-  m = Mutex.new
+  m = Thread::Mutex.new
   Thread.new { m.lock }; m.lock
   :ok
 }
 
 assert_equal 'ok', %q{
-  m = Mutex.new
+  m = Thread::Mutex.new
   Thread.new { m.lock }.join; m.lock
   :ok
 }
 
 assert_equal 'ok', %q{
-  m = Mutex.new
+  m = Thread::Mutex.new
   Thread.new { m.lock; sleep 0.2 }
   sleep 0.1; m.lock
   :ok
 }
 
 assert_equal 'ok', %q{
-  m = Mutex.new
+  m = Thread::Mutex.new
   Thread.new { m.lock; sleep 0.2; m.unlock }
   sleep 0.1; m.lock
   :ok
@@ -406,10 +408,10 @@ assert_equal 'ok', %q{
 }
 
 assert_equal 'ok', %{
-  open("zzz.rb", "w") do |f|
+  open("zzz_t2.rb", "w") do |f|
     f.puts <<-'end;' # do
       begin
-        m = Mutex.new
+        m = Thread::Mutex.new
         parent = Thread.current
         th1 = Thread.new { m.lock; sleep }
         sleep 0.01 until th1.stop?
@@ -430,15 +432,15 @@ assert_equal 'ok', %{
       end
     end;
   end
-  require "./zzz.rb"
+  require "./zzz_t2.rb"
   $result
 }
 
 assert_finish 3, %q{
   require 'thread'
 
-  lock = Mutex.new
-  cond = ConditionVariable.new
+  lock = Thread::Mutex.new
+  cond = Thread::ConditionVariable.new
   t = Thread.new do
     lock.synchronize do
       cond.wait(lock)
@@ -481,4 +483,18 @@ assert_equal 'foo', %q{
   [th1, th2].each {|t| t.join }
   GC.start
   f.call.source
-}
+} unless ENV['RUN_OPTS'] =~ /rjit/ # flaky
+assert_normal_exit %q{
+  class C
+    def inspect
+      sleep 0.5
+      'C!!'
+    end
+  end
+  Thread.new{
+    loop{
+      p C.new
+    }
+  }
+  sleep 0.1
+}, timeout: 5
