@@ -1,48 +1,111 @@
 $LOAD_PATH.unshift File.expand_path('../../lib', __FILE__)
-require 'irb/ruby-lex'
-require 'test/unit'
+require 'irb'
+require 'rubygems'
 require 'ostruct'
 
+require_relative "helper"
+
 module TestIRB
-  class TestRubyLex < Test::Unit::TestCase
+  class TestRubyLex < TestCase
     Row = Struct.new(:content, :current_line_spaces, :new_line_spaces, :nesting_level)
 
     class MockIO_AutoIndent
-      def initialize(params, &assertion)
+      attr_reader :calculated_indent
+
+      def initialize(*params)
         @params = params
-        @assertion = assertion
+        @calculated_indent
       end
 
       def auto_indent(&block)
-        result = block.call(*@params)
-        @assertion.call(result)
+        @calculated_indent = block.call(*@params)
       end
     end
 
-    def assert_indenting(lines, correct_space_count, add_new_line)
+    def setup
+      save_encodings
+    end
+
+    def teardown
+      restore_encodings
+    end
+
+    def calculate_indenting(lines, add_new_line)
       lines = lines + [""] if add_new_line
       last_line_index = lines.length - 1
       byte_pointer = lines.last.length
 
-      ruby_lex = RubyLex.new()
-      io = MockIO_AutoIndent.new([lines, last_line_index, byte_pointer, add_new_line]) do |auto_indent|
-        error_message = "Calculated the wrong number of spaces for:\n #{lines.join("\n")}"
-        assert_equal(correct_space_count, auto_indent, error_message)
-      end
-      ruby_lex.set_input(io)
-      context = OpenStruct.new(auto_indent_mode: true)
-      ruby_lex.set_auto_indent(context)
+      context = build_context
+      context.auto_indent_mode = true
+
+      ruby_lex = RubyLex.new(context)
+      mock_io = MockIO_AutoIndent.new(lines, last_line_index, byte_pointer, add_new_line)
+
+      ruby_lex.set_input(mock_io)
+      ruby_lex.set_auto_indent
+      mock_io.calculated_indent
     end
 
-    def assert_nesting_level(lines, expected)
-      ruby_lex = RubyLex.new()
-      io = proc{ lines.join("\n") }
-      ruby_lex.set_input(io, io)
-      ruby_lex.lex
+    def assert_row_indenting(lines, row)
+      actual_current_line_spaces = calculate_indenting(lines, false)
+
+      error_message = <<~MSG
+        Incorrect spaces calculation for line:
+
+        ```
+      > #{lines.last}
+        ```
+
+        All lines:
+
+        ```
+        #{lines.join("\n")}
+        ```
+      MSG
+      assert_equal(row.current_line_spaces, actual_current_line_spaces, error_message)
+
+      error_message = <<~MSG
+        Incorrect spaces calculation for line after the current line:
+
+        ```
+        #{lines.last}
+      >
+        ```
+
+        All lines:
+
+        ```
+        #{lines.join("\n")}
+        ```
+      MSG
+      actual_next_line_spaces = calculate_indenting(lines, true)
+      assert_equal(row.new_line_spaces, actual_next_line_spaces, error_message)
+    end
+
+    def assert_nesting_level(lines, expected, local_variables: [])
+      ruby_lex = ruby_lex_for_lines(lines, local_variables: local_variables)
       error_message = "Calculated the wrong number of nesting level for:\n #{lines.join("\n")}"
       assert_equal(expected, ruby_lex.instance_variable_get(:@indent), error_message)
     end
 
+    def assert_code_block_open(lines, expected, local_variables: [])
+      ruby_lex = ruby_lex_for_lines(lines, local_variables: local_variables)
+      error_message = "Wrong result of code_block_open for:\n #{lines.join("\n")}"
+      assert_equal(expected, ruby_lex.instance_variable_get(:@code_block_open), error_message)
+    end
+
+    def ruby_lex_for_lines(lines, local_variables: [])
+      context = build_context(local_variables)
+      ruby_lex = RubyLex.new(context)
+
+      io = proc{ lines.join("\n") }
+      ruby_lex.set_input(io) do
+        lines.join("\n")
+      end
+      ruby_lex.lex
+      ruby_lex
+    end
+
     def test_auto_indent
       input_with_correct_indents = [
         Row.new(%q(def each_top_level_statement), nil, 2),
@@ -82,8 +145,7 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
       end
     end
 
@@ -98,8 +160,7 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
       end
     end
 
@@ -117,8 +178,7 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
       end
     end
 
@@ -131,8 +191,7 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
       end
     end
 
@@ -151,15 +210,11 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
       end
     end
 
     def test_endless_range_at_end_of_line
-      if Gem::Version.new(RUBY_VERSION) < Gem::Version.new('2.6.0')
-        pend 'Endless range is available in 2.6.0 or later'
-      end
       input_with_prompt = [
         PromptRow.new('001:0: :> ', %q(a = 3..)),
         PromptRow.new('002:0: :* ', %q()),
@@ -170,6 +225,58 @@ module TestIRB
       assert_dynamic_prompt(lines, expected_prompt_list)
     end
 
+    def test_heredoc_with_embexpr
+      input_with_prompt = [
+        PromptRow.new('001:0:":* ', %q(<<A+%W[#{<<B)),
+        PromptRow.new('002:0:":* ', %q(#{<<C+%W[)),
+        PromptRow.new('003:0:":* ', %q(a)),
+        PromptRow.new('004:0:]:* ', %q(C)),
+        PromptRow.new('005:0:]:* ', %q(a)),
+        PromptRow.new('006:0:":* ', %q(]})),
+        PromptRow.new('007:0:":* ', %q(})),
+        PromptRow.new('008:0:":* ', %q(A)),
+        PromptRow.new('009:0:]:* ', %q(B)),
+        PromptRow.new('010:0:]:* ', %q(})),
+        PromptRow.new('011:0: :> ', %q(])),
+        PromptRow.new('012:0: :* ', %q()),
+      ]
+
+      lines = input_with_prompt.map(&:content)
+      expected_prompt_list = input_with_prompt.map(&:prompt)
+      assert_dynamic_prompt(lines, expected_prompt_list)
+    end
+
+    def test_heredoc_prompt_with_quotes
+      input_with_prompt = [
+        PromptRow.new("001:0:':* ", %q(<<~'A')),
+        PromptRow.new("002:0:':* ", %q(#{foobar})),
+        PromptRow.new("003:0: :> ", %q(A)),
+        PromptRow.new("004:0:`:* ", %q(<<~`A`)),
+        PromptRow.new("005:0:`:* ", %q(whoami)),
+        PromptRow.new("006:0: :> ", %q(A)),
+        PromptRow.new('007:0:":* ', %q(<<~"A")),
+        PromptRow.new('008:0:":* ', %q(foobar)),
+        PromptRow.new('009:0: :> ', %q(A)),
+      ]
+
+      lines = input_with_prompt.map(&:content)
+      expected_prompt_list = input_with_prompt.map(&:prompt)
+      assert_dynamic_prompt(lines, expected_prompt_list)
+    end
+
+    def test_backtick_method
+      input_with_prompt = [
+        PromptRow.new('001:0: :> ', %q(self.`(arg))),
+        PromptRow.new('002:0: :* ', %q()),
+        PromptRow.new('003:0: :> ', %q(def `(); end)),
+        PromptRow.new('004:0: :* ', %q()),
+      ]
+
+      lines = input_with_prompt.map(&:content)
+      expected_prompt_list = input_with_prompt.map(&:prompt)
+      assert_dynamic_prompt(lines, expected_prompt_list)
+    end
+
     def test_incomplete_coding_magic_comment
       input_with_correct_indents = [
         Row.new(%q(#coding:u), nil, 0),
@@ -178,8 +285,7 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
       end
     end
 
@@ -191,8 +297,7 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
       end
     end
 
@@ -204,8 +309,7 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
       end
     end
 
@@ -217,8 +321,7 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
       end
     end
 
@@ -244,8 +347,7 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
       end
     end
 
@@ -275,8 +377,7 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
       end
     end
 
@@ -291,8 +392,7 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
         assert_nesting_level(lines, row.nesting_level)
       end
     end
@@ -312,8 +412,7 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
         assert_nesting_level(lines, row.nesting_level)
       end
     end
@@ -361,8 +460,7 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
         assert_nesting_level(lines, row.nesting_level)
       end
     end
@@ -377,8 +475,7 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
         assert_nesting_level(lines, row.nesting_level)
       end
     end
@@ -393,8 +490,23 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
+        assert_nesting_level(lines, row.nesting_level)
+      end
+    end
+
+    def test_corresponding_syntax_to_keyword_in
+      input_with_correct_indents = [
+        Row.new(%q(module E), nil, 2, 1),
+        Row.new(%q(end), 0, 0, 0),
+        Row.new(%q(class A), nil, 2, 1),
+        Row.new(%q(  in), nil, 4, 1)
+      ]
+
+      lines = []
+      input_with_correct_indents.each do |row|
+        lines << row.content
+        assert_row_indenting(lines, row)
         assert_nesting_level(lines, row.nesting_level)
       end
     end
@@ -409,8 +521,7 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
         assert_nesting_level(lines, row.nesting_level)
       end
     end
@@ -425,8 +536,7 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
         assert_nesting_level(lines, row.nesting_level)
       end
     end
@@ -441,8 +551,7 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
         assert_nesting_level(lines, row.nesting_level)
       end
     end
@@ -457,12 +566,20 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
         assert_nesting_level(lines, row.nesting_level)
       end
     end
 
+    def test_local_variables_dependent_code
+      pend if RUBY_ENGINE == 'truffleruby'
+      lines = ["a /1#/ do", "2"]
+      assert_nesting_level(lines, 1)
+      assert_code_block_open(lines, true)
+      assert_nesting_level(lines, 0, local_variables: ['a'])
+      assert_code_block_open(lines, false, local_variables: ['a'])
+    end
+
     def test_heredoc_with_indent
       input_with_correct_indents = [
         Row.new(%q(<<~Q), 0, 0, 0),
@@ -474,8 +591,7 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
         assert_nesting_level(lines, row.nesting_level)
       end
     end
@@ -491,16 +607,12 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
         assert_nesting_level(lines, row.nesting_level)
       end
     end
 
     def test_broken_heredoc
-      if Gem::Version.new(RUBY_VERSION) < Gem::Version.new('2.7.0')
-        pend 'This test needs Ripper::Lexer#scan to take broken tokens'
-      end
       input_with_correct_indents = [
         Row.new(%q(def foo), nil, 2, 1),
         Row.new(%q(  <<~Q), 2, 2, 1),
@@ -510,8 +622,7 @@ module TestIRB
       lines = []
       input_with_correct_indents.each do |row|
         lines << row.content
-        assert_indenting(lines, row.current_line_spaces, false)
-        assert_indenting(lines, row.new_line_spaces, true)
+        assert_row_indenting(lines, row)
         assert_nesting_level(lines, row.nesting_level)
       end
     end
@@ -532,7 +643,8 @@ module TestIRB
 
     def assert_dynamic_prompt(lines, expected_prompt_list)
       pend if RUBY_ENGINE == 'truffleruby'
-      ruby_lex = RubyLex.new()
+      context = build_context
+      ruby_lex = RubyLex.new(context)
       io = MockIO_DynamicPrompt.new(lines) do |prompt_list|
         error_message = <<~EOM
           Expected dynamic prompt:
@@ -561,6 +673,38 @@ module TestIRB
       assert_dynamic_prompt(lines, expected_prompt_list)
     end
 
+    def test_dyanmic_prompt_with_double_newline_braking_code
+      input_with_prompt = [
+        PromptRow.new('001:1: :* ', %q(if true)),
+        PromptRow.new('002:1: :* ', %q(%)),
+        PromptRow.new('003:1: :* ', %q(;end)),
+        PromptRow.new('004:1: :* ', %q(;hello)),
+        PromptRow.new('005:0: :> ', %q(end)),
+      ]
+
+      lines = input_with_prompt.map(&:content)
+      expected_prompt_list = input_with_prompt.map(&:prompt)
+      assert_dynamic_prompt(lines, expected_prompt_list)
+    end
+
+    def test_dyanmic_prompt_with_multiline_literal
+      input_with_prompt = [
+        PromptRow.new('001:1: :* ', %q(if true)),
+        PromptRow.new('002:1:]:* ', %q(  %w[)),
+        PromptRow.new('003:1:]:* ', %q(  a)),
+        PromptRow.new('004:1: :* ', %q(  ])),
+        PromptRow.new('005:1: :* ', %q(  b)),
+        PromptRow.new('006:1:]:* ', %q(  %w[)),
+        PromptRow.new('007:1:]:* ', %q(  c)),
+        PromptRow.new('008:1: :* ', %q(  ])),
+        PromptRow.new('009:0: :> ', %q(end)),
+      ]
+
+      lines = input_with_prompt.map(&:content)
+      expected_prompt_list = input_with_prompt.map(&:prompt)
+      assert_dynamic_prompt(lines, expected_prompt_list)
+    end
+
     def test_dyanmic_prompt_with_blank_line
       input_with_prompt = [
         PromptRow.new('001:0:]:* ', %q(%w[)),
@@ -574,23 +718,15 @@ module TestIRB
     end
 
     def test_broken_percent_literal
-      if Gem::Version.new(RUBY_VERSION) < Gem::Version.new('2.7.0')
-        pend 'This test needs Ripper::Lexer#scan to take broken tokens'
-      end
-
       tokens = RubyLex.ripper_lex_without_warning('%wwww')
       pos_to_index = {}
       tokens.each_with_index { |t, i|
-        assert_nil(pos_to_index[t[0]], "There is already another token in the position of #{t.inspect}.")
-        pos_to_index[t[0]] = i
+        assert_nil(pos_to_index[t.pos], "There is already another token in the position of #{t.inspect}.")
+        pos_to_index[t.pos] = i
       }
     end
 
     def test_broken_percent_literal_in_method
-      if Gem::Version.new(RUBY_VERSION) < Gem::Version.new('2.7.0')
-        pend 'This test needs Ripper::Lexer#scan to take broken tokens'
-      end
-
       tokens = RubyLex.ripper_lex_without_warning(<<~EOC.chomp)
         def foo
           %wwww
@@ -598,9 +734,108 @@ module TestIRB
       EOC
       pos_to_index = {}
       tokens.each_with_index { |t, i|
-        assert_nil(pos_to_index[t[0]], "There is already another token in the position of #{t.inspect}.")
-        pos_to_index[t[0]] = i
+        assert_nil(pos_to_index[t.pos], "There is already another token in the position of #{t.inspect}.")
+        pos_to_index[t.pos] = i
       }
     end
+
+    def test_unterminated_code
+      ['do', '<<A'].each do |code|
+        tokens = RubyLex.ripper_lex_without_warning(code)
+        assert_equal(code, tokens.map(&:tok).join, "Cannot reconstruct code from tokens")
+        error_tokens = tokens.map(&:event).grep(/error/)
+        assert_empty(error_tokens, 'Error tokens must be ignored if there is corresponding non-error token')
+      end
+    end
+
+    def test_unterminated_heredoc_string_literal
+      context = build_context
+      ['<<A;<<B', "<<A;<<B\n", "%W[\#{<<A;<<B", "%W[\#{<<A;<<B\n"].each do |code|
+        tokens = RubyLex.ripper_lex_without_warning(code)
+        string_literal = RubyLex.new(context).check_string_literal(tokens)
+        assert_equal('<<A', string_literal&.tok)
+      end
+    end
+
+    def test_corresponding_token_depth_with_heredoc_and_embdoc
+      reference_code = <<~EOC.chomp
+        if true
+          hello
+          p(
+          )
+      EOC
+      code_with_heredoc = <<~EOC.chomp
+        if true
+          <<~A
+          A
+          p(
+          )
+      EOC
+      code_with_embdoc = <<~EOC.chomp
+        if true
+        =begin
+        =end
+          p(
+          )
+      EOC
+      context = build_context
+      [reference_code, code_with_heredoc, code_with_embdoc].each do |code|
+        lex = RubyLex.new(context)
+        lines = code.lines
+        lex.instance_variable_set('@tokens', RubyLex.ripper_lex_without_warning(code))
+        assert_equal 2, lex.check_corresponding_token_depth(lines, lines.size)
+      end
+    end
+
+    def test_find_prev_spaces_with_multiline_literal
+      lex = RubyLex.new(build_context)
+      reference_code = <<~EOC.chomp
+        if true
+          1
+          hello
+          1
+          world
+        end
+      EOC
+      code_with_percent_string = <<~EOC.chomp
+        if true
+          %w[
+            hello
+          ]
+          world
+        end
+      EOC
+      code_with_quoted_string = <<~EOC.chomp
+        if true
+          '
+            hello
+          '
+          world
+        end
+      EOC
+      context = build_context
+      [reference_code, code_with_percent_string, code_with_quoted_string].each do |code|
+        lex = RubyLex.new(context)
+        lex.instance_variable_set('@tokens', RubyLex.ripper_lex_without_warning(code))
+        prev_spaces = (1..code.lines.size).map { |index| lex.find_prev_spaces index }
+        assert_equal [0, 2, 2, 2, 2, 0], prev_spaces
+      end
+    end
+
+    private
+
+    def build_context(local_variables = nil)
+      IRB.init_config(nil)
+      workspace = IRB::WorkSpace.new(TOPLEVEL_BINDING.dup)
+
+      if local_variables
+        local_variables.each do |n|
+          workspace.binding.local_variable_set(n, nil)
+        end
+      end
+
+      IRB.conf[:VERBOSE] = false
+      IRB::Context.new(nil, workspace)
+    end
   end
 end
